from os import path
from mock import sentinel
import importlib

from sceptre.context import SceptreContext


class TestSceptreContext(object):

    def setup_method(self, test_method):
        self.context = SceptreContext(
            project_path="project_path/to/sceptre",
            command_path="command/path",
            user_variables={},
            options={},
            output_format="json",
            no_colour=True,
            ignore_dependencies=False
        )

    def test_context_with_path(self):
        context = SceptreContext(
            project_path="project_path/to/sceptre",
            command_path="command-path",
            user_variables=sentinel.user_variables,
            options=sentinel.options,
            output_format=sentinel.output_format,
            no_colour=sentinel.no_colour,
            ignore_dependencies=sentinel.ignore_dependencies
        )

        sentinel.project_path = "project_path/to/sceptre"
        assert context.project_path == sentinel.project_path

    def test_full_config_path_returns_correct_path(self):
        context = SceptreContext(
            project_path="project_path",
            command_path="command-path",
            user_variables=sentinel.user_variables,
            options=sentinel.options,
            output_format=sentinel.output_format,
            no_colour=sentinel.no_colour,
            ignore_dependencies=sentinel.ignore_dependencies
        )

        full_config_path = path.join("project_path", self.context.config_path)
        assert context.full_config_path() == full_config_path

    def test_full_command_path_returns_correct_path(self):
        context = SceptreContext(
            project_path="project_path",
            command_path="command",
            user_variables=sentinel.user_variables,
            options=sentinel.options,
            output_format=sentinel.output_format,
            no_colour=sentinel.no_colour,
            ignore_dependencies=sentinel.ignore_dependencies
        )
        full_command_path = path.join("project_path",
                                      self.context.config_path,
                                      "command")

        assert context.full_command_path() == full_command_path

    def test_full_templates_path_returns_correct_path(self):
        context = SceptreContext(
            project_path="project_path",
            command_path="command",
            user_variables=sentinel.user_variables,
            options=sentinel.options,
            output_format=sentinel.output_format,
            no_colour=sentinel.no_colour,
            ignore_dependencies=sentinel.ignore_dependencies
        )
        full_templates_path = path.join("project_path", self.context.templates_path)
        assert context.full_templates_path() == full_templates_path

    def test_context_repr(self):
        assert self.context.__repr__() == \
            "sceptre.context.SceptreContext(" \
            "project_path='project_path/to/sceptre', " \
            "command_path='command/path', " \
            "user_variables={}, "\
            "options={}, " \
            "output_format='json', " \
            "no_colour=True, " \
            "ignore_dependencies=False" \
            ")"

    def test_repr_can_eval_correctly(self):
        sceptre = importlib.import_module('sceptre')
        mock = importlib.import_module('mock')
        evaluated_context = eval(
            repr(self.context),
            {
                'sceptre': sceptre,
                'sentinel': mock.mock.sentinel
            }
        )
        assert isinstance(evaluated_context, SceptreContext)
        assert evaluated_context.__eq__(self.context)

    def test_context_hash(self):
        assert hash(self.context) == self.context.__hash__()

    def test_command_path_is_stack_with_valid_stack(self):
        self.context.project_path = "tests/fixtures"
        self.context.command_path = "account/stack-group/region/vpc.yaml"
        assert self.context.command_path_is_stack()

    def test_command_path_is_stack_with_directory(self):
        self.context.project_path = "tests/fixtures"
        self.context.command_path = "account/stack-group/region"
        assert self.context.command_path_is_stack() is False



# -*- coding: utf-8 -*-
import abc
import six
import logging

from sceptre.context import SceptreContext
from sceptre.helpers import _call_func_on_values


class ResolverData(object):
    def __init__(self, context):
        if isinstance(context, SceptreContext):
            self.context = context


@six.add_metaclass(abc.ABCMeta)
class Resolver(ResolverData):
    """
    Resolver is an abstract base class that should be inherited by all
    Resolvers.

    :param argument: The argument of the resolver.
    :type argument: str
    :param stack: The associated stack of the resolver.
    :type stack: sceptre.stack.Stack
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, argument=None, stack=None):
        self.logger = logging.getLogger(__name__)
        self.argument = argument
        self.stack = stack

    def setup(self):
        """
        This method is called at during stack initialisation.
        Implementation of this method in subclasses can be used to do any
        initial setup of the object.
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def resolve(self):
        """
        An abstract method which must be overwritten by all inheriting classes.
        This method is called to retrieve the final desired value.
        Implementation of this method in subclasses must return a suitable
        object or primitive type.
        """
        pass  # pragma: no cover


class ResolvableProperty(object):
    """
    This is a descriptor class used to store an attribute that may contain
    Resolver objects. When retrieving the dictionary or list, any Resolver
    objects contains are a value or within a list are resolved to a primitive
    type. Supports nested dictionary and lists.

    :param name: Attribute suffix used to store the property in the instance.
    :type name: str
    """

    def __init__(self, name):
        self.name = "_" + name
        self.logger = logging.getLogger(__name__)

    def __get__(self, instance, type):
        """
        Attribute getter which resolves any Resolver object contained in the
        complex data structure.

        :return: The attribute stored with the suffix ``name`` in the instance.
        :rtype: dict or list
        """
        def resolve(attr, key, value):
            attr[key] = value.resolve()

        if hasattr(instance, self.name):
            return _call_func_on_values(
                resolve, getattr(instance, self.name), Resolver
            )

    def __set__(self, instance, value):
        """
        Attribute setter which adds a stack reference to any resolvers in the
        data structure `value` and calls the setup method.

        """
        def setup(attr, key, value):
            value.stack = instance
            value.setup()

        _call_func_on_values(setup, value, Resolver)
        setattr(instance, self.name, value)



# -*- coding: utf-8 -*-

"""
sceptre.plan.executor

This module implements a SceptrePlanExecutor, which is responsible for
executing the command specified in a SceptrePlan.
"""
import logging

from concurrent.futures import ThreadPoolExecutor, as_completed
from sceptre.plan.actions import StackActions
from sceptre.stack_status import StackStatus


class SceptrePlanExecutor(object):

    def __init__(self, command, launch_order):
        """
        Initalises a SceptrePlanExecutor, generates the launch order, threads
        and intial Stack Statuses.

        :param command: The command to execute on the Stack.
        :type command: str

        :param launch_order: A list containing sets of Stacks that can be\
                executed concurrently.
        :type launch_order: list
        """

        self.logger = logging.getLogger(__name__)
        self.command = command
        self.launch_order = launch_order

        self.num_threads = len(max(launch_order, key=len))
        self.stack_statuses = {stack: StackStatus.PENDING
                               for batch in launch_order for stack in batch}

    def execute(self, *args):
        """
        Execute is responsible executing the sets of Stacks in `launch_order`
        concurrently, in the correct order.

        :param \*args: Any arguments that should be passed through to the\
                StackAction being called.
        """
        responses = {}

        with ThreadPoolExecutor(max_workers=self.num_threads) as executor:
            for batch in self.launch_order:
                futures = [executor.submit(self._execute, stack, *args)
                           for stack in batch]

                for future in as_completed(futures):
                    stack, status = future.result()
                    responses[stack] = status

        return responses

    def _execute(self, stack, *args):
        actions = StackActions(stack)
        result = getattr(actions, self.command)(*args)
        return stack, result
